//
// 	JS Crunch by ssjx (http://ssjx.co.uk)
//	=====================================
//	This program reduces the size of JavaScript files by removing comments and newlines
//	Some files may need editing before the shrunken file will work, this is where the
// 	check option is useful.
//
//	Compiles with either:
//		dmd crunch.d 
//		ldc2 crunch.d
//
// 	Usage: To check a file for errors
//	---------------------------------
//	crunch -check myfile.ja
//
//	Usage: To produce a compressed file using redirection
//	-----------------------------------------------------
//	crunch myfile.js >compress.js
//
//
@safe:

import std.stdio;
import std.string;	//chomp, split, strip
import std.file;

import std.conv;	// to!
	
void main(string[] args){	

	const string ver="v0.1 (21/03/2025)";
	
	enum Actions { NONE, CHECK, COMPRESS, BASIC }

	string fn="";
	auto action=Actions.COMPRESS;
	
	if (args.length>1){
		foreach(par;args){
			if (par[0]=='-'){
				if (par=="-check"){
					action=Actions.CHECK;
					writeln("JS Crunch by ssjx (https://ssjx.co.uk) "~ver);
					writeln("===========================================");
				}
				
				if (par=="-basic"){
					action=Actions.BASIC;
				}
			}else{
				if (par!=args[0]){
					// Not the exe or an option so probably a file
					fn=par;
				}
			}
		}
	}else{
		writeln("JS Crunch by ssjx (https://ssjx.co.uk) "~ver);
		writeln("===========================================");
		writeln("This program can reduce the size of *some* JavaScript files by removing comments");
		writeln("and new line characters.");
		writeln("");
		writeln("Usage : crunch [option] myfile.js");
		writeln("Help:");
		writeln("\tRedirect the output to create the compressed JavaScript file");
		writeln("\te.g. crunch myfile.js >compressed.js");
		writeln("");
		writeln("Options:");
		writeln("\t-check\tLists potential problems (mostly missing semi-colons).");		
		writeln("\t-basic\tNo automatic adding of semi-colons or newlines.");		
		return;
	}
	
	//
	
	if (exists(fn)){	
		auto f = File(fn, "r"); // open for reading
		
		int c=0;
		int no=1;
		int errors=0;
		bool skip=false;
		
		bool strict=false;
		bool usevar=false;
		
		while (!f.eof()){
			string line=f.readln;
			// Chomp removes line feed, strip removes spaces/tabs
			line=line.chomp().strip();
		
			int pos=-1;	// need it to ignore when in quotes
			int slash=0;
			int cskip=0;
			foreach(idx,ch;line){
				if (ch=='\"'){cskip=1-cskip;}
			
				if (cskip==0){
					if (ch=='/'){
						slash++;
						if (slash==2){
							pos=(idx-1);
							break;
						}
					}else{
						slash=0;
					}
				}
			}
			
			if (pos>=0){
				line=line[0..pos];
				line=line.strip();		
			}
			
			if (line.length>0){
				
				if (line.startsWith("/*")){skip=true;}
				
				if (skip==false){
				
					final switch(action){
						case Actions.CHECK:
							bool okay=false;
							
							// Check for let without ;
							if (line.endsWith("}")){okay=true;}
							if (line.endsWith("{")){okay=true;}
							if (line.endsWith(",")){okay=true;}
							if (line.endsWith(";")){okay=true;}
							if (line.endsWith(":")){okay=true;}
							if (line.endsWith("[")){okay=true;}
							if (line.endsWith("]")){okay=true;}		// Not okay but will auto fix!
							if (line.endsWith("\"")){okay=true;}	// Not okay but will auto fix!
							
							if (line.endsWith("*/")){okay=true;}
							if (line.startsWith("switch(") && line.endsWith(")")){okay=true;}
							
							// Check for let without ;
							if (line.startsWith("let ")){
								bool let_okay=false;
								if (line.endsWith(";")==true){let_okay=true;}
								if (line.endsWith("[")==true){let_okay=true;}
								if (line.endsWith("\"")==true){let_okay=true;}
								
								if (let_okay==false){okay=false;}
							}
							
							
							if (okay==false){
								writeln("Possible error line "~to!string(no)~" : "~line);
								errors++;
							}
							
							// Not errors but suggestions
							if (line.startsWith("\"use strict\";")){strict=true;}
							if (line.startsWith("var ") && usevar==false){usevar=true;}
						break;
						
						case Actions.COMPRESS:
							// So we don't add a blank line at the top
							if (c>0){
								bool newline=false;
								if (line.startsWith("function")){newline=true;}	
								if (newline==true){writeln;}
							}
							write(line);
							
							// Add forgotten semi colon
							if (line.endsWith("]")){write(";");}
							if (line.endsWith("\"")){write(";");}
								
							c++;
						break;
						
						case Actions.BASIC:
							write(line);
						break;
						
						case Actions.NONE:
							// Should not get here...
						break;
					}
				}
				
				if (line.endsWith("*/") && skip==true){skip=false;}
			}
			no++;
		} 
		f.close();
		
		// Summary
		if (action==Actions.CHECK){
			writeln(errors.to!string()~" possible error(s) found.");
			
			if (usevar==true || strict==false){
				writeln;
				writeln("Suggestions");
				writeln("===========");
				
				if (strict==false){
					writeln(" *\tCatch problems by adding \"use strict\"; to the top of the document.");
					writeln("\tYou can view errors with the FireFox WebDev Console.");
				}
				
				if (usevar==true){
					writeln(" *\tReplace 'var' with 'let' to add / limit the scope of variables.");
				}
			}
		}
	}else{
		writeln("Could not find the file ",fn);
	}
}
